@@ -1,18 +1,13 @@
 """Italy specific code."""
-from datetime import datetime
 import logging
+from datetime import datetime, timedelta
+
 import requests
 
-from ..const import RESPONSE_OK
 from ..classes.programme import Programme
-
-from .const_it import (
-    CHANNEL_IMAGE_URL,
-    PVR_IMAGE_URL,
-    SCHEDULE_URL,
-    LIVE_IMAGE_URL,
-    CHANNEL_URL,
-)
+from ..const import RESPONSE_OK
+from .const_it import (CHANNEL_IMAGE_URL, CHANNEL_URL, LIVE_IMAGE_URL,
+                       PVR_IMAGE_URL, SCHEDULE_URL)
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -30,15 +25,19 @@ class SkyQCountry:
 
     def getEpgData(self, sid, channelno, epgDate):
         """Get EPG data for Italy."""
-        queryDateFrom = epgDate.strftime("%Y-%m-%dT00:00:00Z")
+        epgPrev = epgDate - timedelta(days=1)
+        queryDateFrom = epgPrev.strftime("%Y-%m-%dT22:00:00Z")
         queryDateTo = epgDate.strftime("%Y-%m-%dT23:59:59Z")
-        return self._getData(sid, channelno, queryDateFrom, queryDateTo)
+        epgData = self._getData(sid, channelno, queryDateFrom, queryDateTo)
 
-    def getTimeEpgData(self, sid, channelno, queryDateFrom, queryDateTo):
-        """Get EPG data for Italy."""
-        queryDateFrom = queryDateFrom.strftime("%Y-%m-%dT%H:%M:%SZ")
-        queryDateTo = queryDateTo.strftime("%Y-%m-%dT%H:%M:%SZ")
-        return self._getData(sid, channelno, queryDateFrom, queryDateTo)
+        midnight = datetime.combine(epgDate.date(), datetime.min.time())
+
+        return [p for p in epgData if p.endtime >= midnight]
+
+    def buildChannelImageUrl(self, sid, channelname):
+        """Build the channel image URL."""
+        chid = "".join(e for e in channelname.casefold() if e.isalnum())
+        return CHANNEL_IMAGE_URL.format(sid, chid)
 
     def _getData(self, sid, channelno, queryDateFrom, queryDateTo):
         cid = None
@@ -94,3 +93,11 @@ class SkyQCountry:
         resp = requests.get(CHANNEL_URL)
         if resp.status_code == RESPONSE_OK:
             self._channellist = resp.json()["channels"]
+
+    def _getMidnight(self, epgDate, epgData):
+        midnight = datetime.combine(epgDate.date(), datetime.min.time())
+        for p in epgData:
+            if p.starttime < midnight and p.endtime >= midnight:
+                return p
+
+        return list(epgData)[len(epgData) - 1]
